"""Constants for the ``paypal_express_checkout`` app."""
from django.conf import settings


# Common values for a payment status.

PAYMENT_STATUS = {
    # app status values
    'checkout': 'Checkout',
    'pending': 'Pending',
    'canceled': 'Canceled',

    # paypal status values
    'completed': 'Completed',
    'canceled_Reversal': 'Canceled_Reversal',
    'created': 'Created',
    'denied': 'Denied',
    'expired': 'Expired',
    'failed': 'Failed',
    'refunded': 'Refunded',
    'reversed': 'Reversed',
    'processed': 'Processed',
    'voided': 'Voided',
}

STATUS_CHOICES = (
    (PAYMENT_STATUS['checkout'], 'Checkout'),
    (PAYMENT_STATUS['pending'], 'Pending'),
    (PAYMENT_STATUS['canceled'], 'Canceled'),
    (PAYMENT_STATUS['completed'], 'Completed'),
    (PAYMENT_STATUS['canceled_Reversal'], 'Canceled_Reversal'),
    (PAYMENT_STATUS['created'], 'Created'),
    (PAYMENT_STATUS['denied'], 'Denied'),
    (PAYMENT_STATUS['expired'], 'Expired'),
    (PAYMENT_STATUS['failed'], 'Failed'),
    (PAYMENT_STATUS['refunded'], 'Refunded'),
    (PAYMENT_STATUS['reversed'], 'Reversed'),
    (PAYMENT_STATUS['processed'], 'Processed'),
    (PAYMENT_STATUS['voided'], 'Voided'),
)


PAYPAL_DEFAULTS = {
    'USER': settings.PAYPAL_USER,
    'PWD': settings.PAYPAL_PWD,
    'SIGNATURE': settings.PAYPAL_SIGNATURE,
    'VERSION': '91.0',
    'PAYMENTREQUEST_0_PAYMENTACTION': 'Sale',
}

if settings.SALE_DESCRIPTION:
    PAYPAL_DEFAULTS.update({
        'PAYMENTREQUEST_0_DESC': settings.SALE_DESCRIPTION})



"""The models for the ``paypal_express_checkout`` app."""
from django.conf import settings
from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import ugettext_lazy as _

from paypal_express_checkout.constants import STATUS_CHOICES


class Item(models.Model):
    """
    Holds the information about an item, that is on Sale.

    The information will be needed to process the PayPal payment transaction.

    :identifier: A unique identifier for the item.
    :name: Name of the item.
    :description: Description of the item.
    :value: The price of the item.
    :currency: Short currency identifier. Defaults to USD.

    """
    identifier = models.CharField(
        max_length=256,
        verbose_name=_('Identifier'),
        blank=True,
    )

    name = models.CharField(
        max_length=2048,
        verbose_name=_('Name'),
    )

    description = models.CharField(
        max_length=4000,
        verbose_name=_('Description'),
    )

    value = models.DecimalField(
        max_digits=8,
        decimal_places=2,
        verbose_name=_('Value'),
    )

    currency = models.CharField(
        max_length=16,
        default='USD',
    )

    def __unicode__(self):
        return '{0} - {1} {2}'.format(self.name, self.value, self.currency)


class PaymentTransaction(models.Model):
    """
    This model holds the information about a payment transaction.

    Needed in the process of the payment as well as later reference.

    :user: The user this transaction is related to.
    :date: The date this transaction has started.
    :transaction_id: The unique identifier of the transaction generated by
      PayPal.
    :value: The amount of the payment. Currency defaults to USD.
    :status: The status of the transaction.

    """
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_('User'),
    )

    content_type = models.ForeignKey(
        ContentType,
        blank=True, null=True,
    )

    object_id = models.PositiveIntegerField(
        blank=True, null=True,
    )

    content_object = generic.GenericForeignKey(
        'content_type',
        'object_id',
    )

    date = models.DateTimeField(
        auto_now=True,
        auto_now_add=True,
        verbose_name=_('Time'),
    )

    transaction_id = models.CharField(
        max_length=32,
        verbose_name=_('Transaction ID'),
    )

    value = models.DecimalField(
        max_digits=8,
        decimal_places=2,
        verbose_name=_('Transaction value'),
    )

    status = models.CharField(
        max_length=16,
        choices=STATUS_CHOICES,
        verbose_name=_('Payment status'),
    )

    class Meta:
        ordering = ['-date', 'transaction_id', ]

    def __unicode__(self):
        return self.transaction_id


class PurchasedItem(models.Model):
    """
    Keeps track of which user purchased which items (and their quantities).

    This helps you to find out if and what your users have purchased.

    :user: FK to the user who made the purchase.
    :identifier: An identifier to select items of the same type. This can be
      helpful if you want to calculate the total shipping costs and the total
      cost of goods for a certain transaction.
    :transaction: The transaction that belongs to this purchase
    :content_object: Use this if you would like to point at any kind of item
      and not the Item model of this app.
    :item: The item that belongs to this purchase
    :price: The price of the item at the time of the purchase
    :quantity: The quantity of items that has been purchased

    """
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_('User'),
    )

    identifier = models.CharField(
        max_length=256,
        verbose_name=_('Identifier'),
        blank=True,
    )

    transaction = models.ForeignKey(
        PaymentTransaction,
        verbose_name=_('Transaction'),
    )

    item = models.ForeignKey(
        Item,
        verbose_name=_('Item'),
        null=True, blank=True,
    )

    content_type = models.ForeignKey(
        ContentType,
        blank=True, null=True,
    )

    object_id = models.PositiveIntegerField(
        blank=True, null=True,
    )

    content_object = generic.GenericForeignKey(
        'content_type',
        'object_id',
    )

    price = models.FloatField(
        verbose_name=_('Price'),
        blank=True, null=True,
    )

    quantity = models.PositiveIntegerField(
        verbose_name=_('Quantity'),
    )

    class Meta:
        ordering = ['-transaction__date', 'transaction__transaction_id', ]

    def __unicode__(self):
        return '{0} {1} of {2} [{3}]'.format(
            self.quantity, self.item, self.user.email, self.transaction)


class PaymentTransactionError(models.Model):
    """
    A model to track errors during payment process.

    :data: When the error ocurred.
    :user: For which user the error occurred.
    :paypal_api_url: The API endpoint we have been calling, which has responded
      with the error.
    :request_data: The data payload we have been sending to the API endpoint.
    :response: The full response string from PayPal.
    :transaction: If we send a request at a point in time where we already have
      a transaction, we can add a FK to that transaction for easier cross
      referencing.

    """
    date = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_('Time'),
    )

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_('User'),
    )

    paypal_api_url = models.CharField(
        max_length=4000,
        verbose_name=_('Paypal API URL'),
        blank=True,
    )

    request_data = models.TextField(
        verbose_name=_('Request data'),
        blank=True,
    )

    response = models.TextField(
        verbose_name=_('Response String'),
        blank=True,
    )

    transaction = models.ForeignKey(
        PaymentTransaction,
        blank=True, null=True,
        verbose_name=_('Payment transaction'),
    ),



"""Forms for the ``paypal_express_checkout`` app."""
import httplib
import logging
import urllib
import urllib2
import urlparse

from django import forms
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.urlresolvers import reverse
from django.http import Http404
from django.shortcuts import redirect
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from .constants import PAYMENT_STATUS, PAYPAL_DEFAULTS
from .models import (
    Item,
    PaymentTransaction,
    PaymentTransactionError,
    PurchasedItem,
)
from .settings import API_URL, LOGIN_URL
from .utils import urlencode


logger = logging.getLogger(__name__)


CURRENCYCODE = getattr(settings, 'PAYPAL_CURRENCYCODE', 'USD')


class PayPalFormMixin(object):
    """Common methods for the PayPal forms."""
    def call_paypal(self, api_url, post_data, transaction=None):
        """
        Gets the PayPal API URL from the settings and posts ``post_data``.

        :param api_url: The API endpoint that should be called.
        :param post_data: The full post data for PayPal containing all needed
          information for the current transaction step.
        :param transaction: If you already have a transaction, pass it into
          this method so that it can be logged in case of an error.

        """
        data = urlencode(post_data)
        try:
            response = urllib2.urlopen(api_url, data=data)
        except (
                urllib2.HTTPError,
                urllib2.URLError,
                httplib.HTTPException), ex:
            self.log_error(
                ex, api_url=api_url, request_data=data,
                transaction=transaction)
        else:
            parsed_response = urlparse.parse_qs(response.read())
            return parsed_response

    def get_cancel_url(self):
        """Returns the paypal cancel url."""
        return settings.HOSTNAME + reverse(
            'paypal_canceled', kwargs=self.get_url_kwargs())

    def get_error_url(self):
        """Returns the url of the payment error page."""
        return reverse('paypal_error')

    def get_notify_url(self):
        """Returns the notification (ipn) url."""
        return settings.HOSTNAME + reverse('ipn_listener')

    def get_return_url(self):
        """Returns the paypal return url."""
        return settings.HOSTNAME + reverse(
            'paypal_confirm', kwargs=self.get_url_kwargs())

    def get_success_url(self):
        """Returns the url of the payment success page."""
        return reverse('paypal_success')

    def log_error(self, error_message, api_url=None, request_data=None,
                  transaction=None):
        """
        Saves error information as a ``PaymentTransactionError`` object.

        :param error_message: The message of the exception or response string
          from PayPal.

        """
        payment_error = PaymentTransactionError()
        payment_error.user = self.user
        payment_error.response = error_message
        payment_error.paypal_api_url = api_url
        payment_error.request_data = request_data
        payment_error.transaction = transaction
        payment_error.save()
        return payment_error


class DoExpressCheckoutForm(PayPalFormMixin, forms.Form):
    """
    Takes the input from the ``DoExpressCheckoutView``, validates it and
    takes care of the PayPal API operations.

    """
    token = forms.CharField()

    PayerID = forms.CharField()

    def __init__(self, user, *args, **kwargs):
        self.user = user
        super(DoExpressCheckoutForm, self).__init__(*args, **kwargs)
        try:
            self.transaction = PaymentTransaction.objects.get(
                user=user, transaction_id=self.data['token'])
        except PaymentTransaction.DoesNotExist:
            raise Http404

    def get_post_data(self):
        """Creates the post data dictionary to send to PayPal."""
        post_data = PAYPAL_DEFAULTS.copy()
        items = self.transaction.purchaseditem_set.all()
        if len(items) != 0 and getattr(
                items[0].item, 'currency', None) is not None:
            currency = items[0].item.currency
        else:
            currency = CURRENCYCODE
        post_data.update({
            'METHOD': 'DoExpressCheckoutPayment',
            'TOKEN': self.transaction.transaction_id,
            'PAYERID': self.data['PayerID'],
            'PAYMENTREQUEST_0_AMT': self.transaction.value,
            'PAYMENTREQUEST_0_NOTIFYURL': self.get_notify_url(),
            'PAYMENTREQUEST_0_CURRENCYCODE': currency,
        })
        return post_data

    def do_checkout(self):
        """Calls PayPal to make the 'DoExpressCheckoutPayment' procedure."""
        post_data = self.get_post_data()
        api_url = API_URL
        parsed_response = self.call_paypal(api_url, post_data)
        if parsed_response.get('ACK')[0] == 'Success':
            transaction_id = parsed_response.get(
                'PAYMENTINFO_0_TRANSACTIONID')[0]
            self.transaction.transaction_id = transaction_id
            self.transaction.status = PAYMENT_STATUS['pending']
            self.transaction.save()
            return redirect(self.get_success_url())
        elif parsed_response.get('ACK')[0] == 'Failure':
            self.transaction.status = PAYMENT_STATUS['canceled']
            self.transaction.save()
            # we have to do urlencode here to make the post data more readable
            # in the error log
            post_data_encoded = urlencode(post_data)
            self.log_error(
                parsed_response, api_url, request_data=post_data_encoded,
                transaction=self.transaction)
            return redirect(self.get_error_url())


class SetExpressCheckoutFormMixin(PayPalFormMixin, forms.Form):
    """
    Base form class for all forms invoking the ``SetExpressCheckout`` PayPal
    API operation, providing the general method skeleton.

    Also this is to be used to construct custom forms.

    :param user: The user making the purchase
    :param redirect: If ``True``, the form will return a HttpResponseRedirect,
      otherwise it will only return the redirect URL. This can be useful if
      you want to use this form in an AJAX view.

    """
    def __init__(self, user, redirect=True, *args, **kwargs):
        self.redirect = redirect
        self.user = user
        super(SetExpressCheckoutFormMixin, self).__init__(*args, **kwargs)

    def get_content_object(self):
        """
        Can be overridden to return a different content object for the
        PaymentTransaction model.

        This is useful if you want e.g. have one of your models assigned to the
        transaction for easier identification.

        """
        # TODO for now it should return the user, although I know, that the
        # user is already present in the user field of the PaymentTransaction
        # model.
        # Maybe we can remove the user field safely in exchange for the generic
        # relation only.
        return self.user

    def get_item(self):
        """Obsolete. Just implement ``get_items_and_quantities``."""
        raise NotImplementedError

    def get_quantity(self):
        """Obsolete. Just implement ``get_items_and_quantities``."""
        raise NotImplementedError

    def get_items_and_quantities(self):
        """
        Returns the items and quantities and content objects.

        Content objects are optional, return None if you don't need it.

        Should return a list of tuples:
        ``[(item, quantity, content_object), ]``

        """
        logger.warning(
            'Deprecation warning: Please implement get_items_and_quantities on'
            ' your SetExpressCheckoutForm. Do not use get_item and'
            ' get_quantity any more.')
        return [(self.get_item(), self.get_quantity(), None), ]

    def get_post_data(self, item_quantity_list):
        """Creates the post data dictionary to send to PayPal."""
        post_data = PAYPAL_DEFAULTS.copy()
        total_value = 0
        item_index = 0
        for item, quantity, content_type in item_quantity_list:
            if not quantity:
                # If a user chose quantity 0, we don't include it
                continue
            total_value += item.value * quantity
            post_data.update({
                'L_PAYMENTREQUEST_0_NAME{0}'.format(
                    item_index): item.name,
                'L_PAYMENTREQUEST_0_DESC{0}'.format(
                    item_index): item.description,
                'L_PAYMENTREQUEST_0_AMT{0}'.format(
                    item_index): item.value,
                'L_PAYMENTREQUEST_0_QTY{0}'.format(
                    item_index): quantity,
            })
            item_index += 1

        if (
                len(item_quantity_list) != 0 and
                getattr(item_quantity_list[0][0], 'currency') is not None):
            currency = item_quantity_list[0][0].currency
        else:
            currency = CURRENCYCODE

        post_data.update({
            'METHOD': 'SetExpressCheckout',
            'PAYMENTREQUEST_0_AMT': total_value,
            'PAYMENTREQUEST_0_ITEMAMT': total_value,
            'RETURNURL': self.get_return_url(),
            'CANCELURL': self.get_cancel_url(),
            'PAYMENTREQUEST_0_CURRENCYCODE': currency,
        })
        return post_data

    def get_url_kwargs(self):
        """Provide additional url kwargs, by overriding this method."""
        return {}

    def post_transaction_save(self, transaction, item_quantity_list):
        """
        Override this method if you need to create further objects.

        Once we got a successful response from PayPal we can create a
        Transaction with status "checkout". You might want to create or
        manipulate further objects in your app at this point.

        For example you might ask for user's the t-shirt size on your checkout
        form. This a good place to save the user's choice on the UserProfile.

        """
        return

    def set_checkout(self):
        """
        Calls PayPal to make the 'SetExpressCheckout' procedure.

        :param items: A list of ``Item`` objects.

        """
        item_quantity_list = self.get_items_and_quantities()
        post_data = self.get_post_data(item_quantity_list)
        api_url = API_URL

        # making the post to paypal and handling the results
        parsed_response = self.call_paypal(api_url, post_data)
        if parsed_response.get('ACK')[0] == 'Success':
            token = parsed_response.get('TOKEN')[0]
            transaction = PaymentTransaction(
                user=self.user,
                date=now(),
                transaction_id=token,
                value=post_data['PAYMENTREQUEST_0_AMT'],
                status=PAYMENT_STATUS['checkout'],
                content_object=self.get_content_object(),
            )
            transaction.save()
            self.post_transaction_save(transaction, item_quantity_list)
            for item, quantity, content_object in item_quantity_list:
                if not quantity:
                    continue
                purchased_item_kwargs = {
                    'user': self.user,
                    'transaction': transaction,
                    'quantity': quantity,
                    'price': item.value,
                    'identifier': item.identifier,
                }

                if content_object:
                    purchased_item_kwargs.update({
                        'object_id': content_object.pk,
                        'content_type': ContentType.objects.get_for_model(
                            content_object),
                    })

                if item.pk:
                    purchased_item_kwargs.update({'item': item, })

                PurchasedItem.objects.create(**purchased_item_kwargs)
            if self.redirect:
                return redirect(LOGIN_URL + token)
            return LOGIN_URL + token
        elif parsed_response.get('ACK')[0] == 'Failure':
            post_data_encoded = urlencode(post_data)
            self.log_error(
                parsed_response, api_url=api_url,
                request_data=post_data_encoded)
            return redirect(self.get_error_url())


class SetExpressCheckoutItemForm(SetExpressCheckoutFormMixin):
    """
    Takes the input from the ``SetExpressCheckoutView``, validates it and
    takes care of the PayPal API operations.

    """
    item = forms.ModelChoiceField(
        queryset=Item.objects.all(),
        empty_label=None,
        label=_('Item'),
    )

    quantity = forms.IntegerField(
        label=_('Quantity'),
    )

    def get_item(self):
        """Keeping this for backwards compatibility."""
        return self.cleaned_data.get('item')

    def get_quantity(self):
        """Keeping this for backwards compatibility."""
        return self.cleaned_data.get('quantity')

    def get_items_and_quantities(self):
        """
        Returns the items and quantities.

        Should return a list of tuples.

        """
        return [
            (self.get_item(), self.get_quantity(), None),
        ]



"""Settings for the ``paypal_express_checkout`` app."""
from django.conf import settings


SET_CHECKOUT_FORM = getattr(
    settings, 'PAYPAL_SET_CHECKOUT_FORM',
    'paypal_express_checkout.forms.SetExpressCheckoutItemForm')

LOGIN_URL = getattr(
    settings, 'PAYPAL_LOGIN_URL',
    'https://www.paypal.com/cgi-bin/webscr?cmd=_express-checkout&token=')

API_URL = getattr(
    settings, 'PAYPAL_API_URL',
    'https://api.paypal.com/nvp')

ALLOW_ANONYMOUS_CHECKOUT = getattr(
    settings, 'PAYPAL_ALLOW_ANONYMOUS_CHECKOUT',
    False)



"""Factories for the ``paypal_express_checkout`` app."""
import factory
from decimal import Decimal

from django_libs.tests.factories import UserFactory

from .. import models


class ItemFactory(factory.DjangoModelFactory):
    FACTORY_FOR = models.Item

    name = factory.Sequence(lambda x: 'item{0}'.format(x))
    description = factory.Sequence(lambda x: 'item descriptio{0}'.format(x))
    value = Decimal('10.00')


class PaymentTransactionFactory(factory.DjangoModelFactory):
    FACTORY_FOR = models.PaymentTransaction

    user = factory.SubFactory(UserFactory)
    transaction_id = factory.Sequence(lambda x: '123abc{0}'.format(x))
    value = Decimal('10.00')


class PurchasedItemFactory(factory.DjangoModelFactory):
    FACTORY_FOR = models.PurchasedItem

    user = factory.SubFactory(UserFactory)
    transaction = factory.LazyAttribute(
        lambda x: PaymentTransactionFactory(user=x.user))
    item = factory.SubFactory(ItemFactory)
    quantity = 1



"""Tests for the views of the ``paypal_express_checkout`` app."""
from mock import Mock, patch

from django.conf import settings
from django.core.urlresolvers import reverse
from django.test import TestCase, RequestFactory

from django_libs.tests.factories import UserFactory
from django_libs.tests.mixins import ViewTestMixin

from ..factories import (
    ItemFactory,
    PaymentTransactionFactory,
)
from ...forms import PayPalFormMixin
from ...signals import payment_completed
from ...views import DoExpressCheckoutView


class PaymentViewTestCaseMixin(ViewTestMixin):
    def should_redirect_to_login_when_anonymous(self):
        """Custom method to overwrite the one from django_libs."""
        url = self.get_url()
        resp = self.client.get(url)
        self.assertRedirects(resp, '{0}?next={1}'.format('/', url))


class DoExpressCheckoutViewTestCase(PaymentViewTestCaseMixin, TestCase):
    """Tests for the ``DoExpressCheckoutView`` view class."""
    longMessage = True

    def get_view_name(self):
        return 'paypal_confirm'

    def get_post_data(self):
        return {
            'token': self.transaction.transaction_id,
            'payerID': 'testpayerID',
        }

    def setUp(self):
        self.user = UserFactory()
        self.transaction = PaymentTransactionFactory(user=self.user)
        self.item = ItemFactory()

        self.paypal_response = {
            'ACK': ['Success'],
            'TOKEN': ['abc123'],
            'PAYMENTINFO_0_TRANSACTIONID': ['abc123']
        }

        self.data = {
            'item': self.item.pk,
            'quantity': 1,
        }

        self.old_call_paypal = PayPalFormMixin.call_paypal
        PayPalFormMixin.call_paypal = Mock(return_value=self.paypal_response)

    def tearDown(self):
        PayPalFormMixin.call_paypal = self.old_call_paypal

    def get_req_and_view(self):
        self.login(self.user)
        req = RequestFactory().get('/', data=self.get_post_data())
        req.user = self.user
        view = DoExpressCheckoutView()
        return req, view

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.is_callable(user=self.user, data=self.get_post_data())
        self.is_callable('post', data=self.get_post_data())
        self.is_not_callable()

    def test_skip_confirmation(self):
        req, view = self.get_req_and_view()
        view.skip_confirmation = True
        view.post = Mock()
        view.dispatch(req)
        self.assertEqual(view.post.call_count, 1, msg=(
            'When called with skip_confirmation=True, the view will not'
            ' render the template but execute the post handler instead'
            ' of the get handler'))

    def test_form_valid(self):
        req, view = self.get_req_and_view()
        form_mock = Mock()
        view.form = form_mock
        view.dispatch(req)
        view.form_valid(view.form)
        self.assertEqual(form_mock.do_checkout.call_count, 1, msg=(
            'When the form is valid, the view should execute and return the'
            ' do_checkout method of the form'))


class PaymentCancelViewTestCase(PaymentViewTestCaseMixin, TestCase):
    """Tests for the ``PaymentCancelView`` view class."""
    longMessage = True

    def get_view_name(self):
        return 'paypal_canceled'

    def setUp(self):
        self.user = UserFactory()

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.should_be_callable_when_authenticated(self.user)


class PaymentErrorViewTestCase(PaymentViewTestCaseMixin, TestCase):
    """Tests for the ``PaymentErrorView`` view class."""
    longMessage = True

    def get_view_name(self):
        return 'paypal_error'

    def setUp(self):
        self.user = UserFactory()

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.should_be_callable_when_authenticated(self.user)


class PaymentSuccessViewTestCase(PaymentViewTestCaseMixin, TestCase):
    """Tests for the ``PaymentSuccessView`` view class."""
    longMessage = True

    def get_view_name(self):
        return 'paypal_success'

    def setUp(self):
        self.user = UserFactory()

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.should_be_callable_when_authenticated(self.user)


class SetExpressCheckoutViewTestCase(PaymentViewTestCaseMixin, TestCase):
    """Tests for the ``SetExpressCheckoutView`` view class."""
    longMessage = True

    def get_view_name(self):
        return 'paypal_checkout'

    def get_post_data(self):
        return {
            'USER': settings.PAYPAL_USER,
            'PWD': settings.PAYPAL_PWD,
            'SIGNATURE': settings.PAYPAL_SIGNATURE,
            'VERSION': '91.0',
            'PAYMENTREQUEST_0_PAYMENTACTION': 'Sale',
            'METHOD': 'SetExpressCheckout',

            'L_PAYMENTREQUEST_0_NAME0': self.item.name,
            'L_PAYMENTREQUEST_0_DESC0': self.item.description,
            'L_PAYMENTREQUEST_0_AMT0': self.item.value,
            'L_PAYMENTREQUEST_0_QTY0': 1,

            'PAYMENTREQUEST_0_AMT': self.item.value,
            'PAYMENTREQUEST_0_ITEMAMT': self.item.value,
            'RETURNURL': settings.HOSTNAME + reverse(
                'paypal_confirm'),
            'CANCELURL': settings.HOSTNAME + reverse(
                'paypal_canceled')
        }

    def setUp(self):
        self.user = UserFactory()
        self.item = ItemFactory()

        self.paypal_response = {
            'ACK': ['Success'],
            'TOKEN': ['abc123'],
        }

        self.data = {
            'item': self.item.pk,
            'quantity': 1,
        }

    @patch.object(PayPalFormMixin, 'call_paypal')
    def test_view(self, call_paypal_mock):
        call_paypal_mock.return_value = self.paypal_response
        self.should_redirect_to_login_when_anonymous()
        self.is_callable('post', data=self.data, user=self.user)


class IPNListenerViewTestCase(ViewTestMixin, TestCase):
    """Tests for the ``IPNListenerView`` view class."""
    longMessage = True

    def get_view_name(self):
        return 'ipn_listener'

    def receive(self, signal, sender, transaction):
        self.signal = signal
        self.sender = sender
        self.received_transaction = transaction

    def setUp(self):
        self.transaction = PaymentTransactionFactory()
        self.valid_data = {
            'txn_id': self.transaction.transaction_id,
            'payment_status': 'Completed',
        }
        self.ipn_received = False
        self.signal = payment_completed.connect(self.receive)

    def test_is_callable_and_sends_signal(self):
        self.is_callable(method='post', data=self.valid_data)

        self.assertEqual(self.received_transaction, self.transaction, msg=(
            'When the IPNListenerView is called, it should send a signal.'))

        self.is_not_callable()



"""Admins for the models of the ``paypal_express_checkout`` app."""
import django
from django.contrib import admin

from . import models


try:
    user_model = django.contrib.auth.get_user_model()
except AttributeError:
    user_model = django.contrib.auth.models.User

username_field = getattr(user_model, 'USERNAME_FIELD', 'username')


class ItemAdmin(admin.ModelAdmin):
    """Custom admin for the ``Item`` model."""
    list_display = ['name', 'description_short', 'value']
    search_fields = ['name', 'description']

    def description_short(self, obj):
        return '{0}...'.format(obj.description[:50])


class PaymentTransactionAdmin(admin.ModelAdmin):
    """Custom admin for the ``PaymentTransaction`` model."""
    list_display = [
        'date', 'user', 'user_email', 'transaction_id', 'value', 'status']
    search_fields = [
        'transaction_id', 'status', 'user__email', 'user__' + username_field]
    list_filter = ['status']
    raw_id_fields = ['user', ]

    def user_email(self, obj):
        return obj.user.email


class PaymentTransactionErrorAdmin(admin.ModelAdmin):
    """Custom admin for the ``PaymentTransactionError`` model."""
    list_display = [
        # FIXME 'transaction_id'
        'date', 'user', 'user_email', 'response_short',
    ]

    def user_email(self, obj):
        return obj.user.email

    def response_short(self, obj):
        return '{0}...'.format(obj.response[:50])

    def transaction_id(self, obj):
        return obj.transaction_id


class PurchasedItemAdmin(admin.ModelAdmin):
    """Custom admin for the ``PurchasedItem`` model."""
    list_display = [
        'date', 'user', 'user_email', 'transaction', 'item', 'price',
        'quantity', 'subtotal', 'total', 'status', ]
    list_filter = [
        'transaction__status', 'item', ]
    search_fields = [
        'transaction__transaction_id', 'user__email', ]
    raw_id_fields = ['user', 'transaction', ]

    def date(self, obj):
        return obj.transaction.date

    def status(self, obj):
        return obj.transaction.status

    def subtotal(self, obj):
        price = 0
        if obj.item is not None:
            price = obj.item.value
        if obj.price:
            price = obj.price
        return price * obj.quantity

    def total(self, obj):
        return obj.transaction.value

    def user_email(self, obj):
        return obj.user.email


admin.site.register(models.Item, ItemAdmin)
admin.site.register(models.PaymentTransaction, PaymentTransactionAdmin)
admin.site.register(
    models.PaymentTransactionError, PaymentTransactionErrorAdmin)
admin.site.register(models.PurchasedItem, PurchasedItemAdmin)



"""Settings that need to be set in order to run the tests."""
import os

DEBUG = True

SITE_ID = 1

AUTH_USER_MODEL = 'auth.User'

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

ROOT_URLCONF = 'paypal_express_checkout.tests.urls'

STATIC_URL = '/static/'

STATIC_ROOT = os.path.join(__file__, '../../../static/')

STATICFILES_DIRS = (
    os.path.join(__file__, 'test_static'),
)

TEMPLATE_DIRS = (
    os.path.join(os.path.dirname(__file__), '../templates'),
)

COVERAGE_REPORT_HTML_OUTPUT_DIR = os.path.join(
    os.path.dirname(__file__), 'coverage')

COVERAGE_MODULE_EXCLUDES = [
    'tests$', 'settings$', 'urls$', 'locale$',
    'migrations', 'fixtures', 'admin$', 'django_extensions', '__init__',
]

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
)


EXTERNAL_APPS = [
    'django.contrib.admin',
    'django.contrib.admindocs',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
    'django.contrib.sitemaps',
    'django.contrib.sites',
    'django_libs',
]

INTERNAL_APPS = [
    'django_nose',
    'paypal_express_checkout',
]

INSTALLED_APPS = EXTERNAL_APPS + INTERNAL_APPS

COVERAGE_MODULE_EXCLUDES += EXTERNAL_APPS

# this leads to the dummy home page added for testing purposes.
LOGIN_URL = '/'


# ======================
# necessary app settings
# ======================
HOSTNAME = 'http://localhost:8000'  # without trailing slash

PAYPAL_API_URL = 'https://api.sandbox.paypal.com/nvp'

PAYPAL_LOGIN_URL = (
    'https://www.sandbox.paypal.com/cgi-bin/'
    'webscr?cmd=_express-checkout&token='
)

# [OPTIONAL]
# The general description of all Sales (not the one of a single item!)
SALE_DESCRIPTION = 'Your payment to {0}'.format(HOSTNAME)


# Settings for local_settings.py
PAYPAL_USER = 'api_user@example.com'
PAYPAL_PWD = 'your api password'
PAYPAL_SIGNATURE = 'your api signature'



"""Views for the ``paypal_express_checkout`` app."""
from django.contrib.auth.decorators import login_required
from django.http import Http404, HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import FormView, TemplateView, View
from django.utils.decorators import method_decorator

from django_libs.utils import conditional_decorator

from . import settings
from .constants import PAYMENT_STATUS
from .forms import (
    DoExpressCheckoutForm,
)
from .models import PaymentTransaction
from .signals import payment_completed, payment_status_updated
from .settings import SET_CHECKOUT_FORM


# importing the name of the current form for SetExpressCheckout
class_name = SET_CHECKOUT_FORM.split('.')[-1]
module_name = '.'.join(SET_CHECKOUT_FORM.split('.')[:-1])
module = __import__(module_name, fromlist=[class_name])
SetExpressCheckoutForm = getattr(module, class_name)


class PaymentViewMixin(object):
    """A Mixin to combine common methods of several payment related views."""
    @conditional_decorator(
        method_decorator(login_required),
        not settings.ALLOW_ANONYMOUS_CHECKOUT)
    def dispatch(self, request, *args, **kwargs):
        """Makes sure that the user is logged in."""
        if not settings.ALLOW_ANONYMOUS_CHECKOUT:
            self.user = request.user
        return super(PaymentViewMixin, self).dispatch(request, *args, **kwargs)


class DoExpressCheckoutView(PaymentViewMixin, FormView):
    """
    This view lets the user review and confirm the payment.

    It leads to the ``DoExpressCheckout`` PayPal API operation.

    :attr skip_confirmation: If you set this to ``True``, the user will not
      need to click yet another confirm button, instead, the paypal website
      will load a little bit longer and we will get straight to the success
      view.

    """
    form_class = DoExpressCheckoutForm
    template_name = 'paypal_express_checkout/confirm_checkout.html'
    skip_confirmation = False

    @conditional_decorator(
        method_decorator(login_required),
        not settings.ALLOW_ANONYMOUS_CHECKOUT)
    def dispatch(self, request, *args, **kwargs):
        """Recalls the transaction using the paypal token."""
        # when this view posts to itself it sends the info in the post data
        self.token = request.GET.get('token') or request.POST.get('token')
        self.payerID = request.GET.get('PayerID') or request.POST.get(
            'PayerID')
        try:
            self.transaction = PaymentTransaction.objects.get(
                user=request.user, transaction_id=self.token)
        except PaymentTransaction.DoesNotExist:
            raise Http404

        if self.skip_confirmation:
            self.user = request.user
            self.request = request
            return self.post(request, *args, **kwargs)
        return super(DoExpressCheckoutView, self).dispatch(
            request, *args, **kwargs)

    def form_valid(self, form):
        """When the form is valid, the form should handle the PayPal call."""
        return form.do_checkout()

    def get_context_data(self, **kwargs):
        ctx = super(DoExpressCheckoutView, self).get_context_data(**kwargs)
        ctx.update({
            'value': self.transaction.value,
            'token': self.token,
            'payerid': self.payerID,
        })
        return ctx

    def get_form_kwargs(self):
        kwargs = super(DoExpressCheckoutView, self).get_form_kwargs()
        kwargs.update({'user': self.user})
        # PayPal makes a GET request with the data, so we check if the GET data
        # is populated and overwrite form data with it.
        if any(self.request.GET):
            kwargs.update({'data': self.request.GET})
        return kwargs


class PaymentCancelView(PaymentViewMixin, TemplateView):
    """The user is redirected to this view, if he cancels the payment."""
    template_name = 'paypal_express_checkout/payment_canceled.html'


class PaymentErrorView(PaymentViewMixin, TemplateView):
    """The user is redirected here once an error occurs during payment."""
    template_name = 'paypal_express_checkout/error.html'


class PaymentSuccessView(PaymentViewMixin, TemplateView):
    """The user is redirected here after a successful PayPal payment."""
    template_name = 'paypal_express_checkout/success.html'


class SetExpressCheckoutView(PaymentViewMixin, FormView):
    """
    This view lets the user initiate the payment.

    It leads to the ``SetExpressCheckout`` PayPal API operation.

    """
    form_class = SetExpressCheckoutForm
    template_name = 'paypal_express_checkout/set_checkout.html'
    redirect = True

    def form_valid(self, form):
        """When the form is valid, the form should handle the PayPal call."""
        return form.set_checkout()

    def get_form_kwargs(self):
        kwargs = super(SetExpressCheckoutView, self).get_form_kwargs()
        kwargs.update({'user': self.user})
        return kwargs


class IPNListenerView(View):
    """This view handles an IPN from PayPal."""
    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        transaction_id = request.POST.get('txn_id')
        try:
            self.payment_transaction = PaymentTransaction.objects.get(
                transaction_id=transaction_id)
        except PaymentTransaction.DoesNotExist:
            raise Http404
        return super(IPNListenerView, self).dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        payment_status = request.POST.get('payment_status')
        self.payment_transaction.status = payment_status
        self.payment_transaction.save()
        if payment_status == PAYMENT_STATUS['completed']:
            payment_completed.send(self, transaction=self.payment_transaction)
        payment_status_updated.send(self, transaction=self.payment_transaction)
        return HttpResponse()



"""Signals for the ``paypal_express_checkout`` app."""
from django.dispatch import Signal


# A signal that arrives, when an IPN is received with status 'Completed'
payment_completed = Signal(providing_args=['transaction'])
payment_status_updated = Signal(providing_args=['transaction'])



"""Utilities for the paypal_express_checkout app."""
import urllib


def urlencode(data):
    for key, value in data.iteritems():
        data[key] = unicode(value).encode('utf-8')
    return urllib.urlencode(data)
